/*
	This program compares two files with identical lengths and generates a
	binary patch file which contains the differences between the two files.
	The binary patch file may then be executed against one of the two files
	to convert it into the other file (not very well explained but I'm sure you
	get the idea.

	The patch file produced will contain a number of records which specify the
	file differences.  The format for the patch file is as follows:

	<file length>		long - this is the length of the file to which the
							   patch may be applied safely.
	<sequence of patch records>

	The definition of a patch record is given below.

	<fileoffset>		long - offset of start of differences
	<Number of bytes>	long - number of differences in the file found starting
							   at this offset.
	<original byte>		char - The value at this position in the original file.
	<new byte>			char - The value at this position in the new file.

	The program should be used in the following manner

	MAKPATCH <file 1> <file 2> <patch file> -(v || q)

	The -v flag puts the patch program in verbose mode.  The -q flag puts the
	program in quiet mode.  Quiet mode is the default.

	Date		By	Description
	-----------	---	-------------------------------------------------------
	16-May-1998	DML	At long last I have a chance to actually sit down and write
					this thing.  The original code is taken from a utility I
					write a long long time ago, in what seems now like another
					life.  I actually had a life in those days.
*/
#include <stdio.h>

/*
	This function determines the length of a file
*/
long filesize(FILE *fFile)
{
	long	lCurrentPosition;
	long	lLength;

	//	Store the current position within the file so we can restore it after
	//	determining the length of the file.
	lCurrentPosition = ftell( fFile );

	//	Determine the length of the file by moving to the end of the file
	//	and asking for the current file offset.
	fseek( fFile, 0L, SEEK_END );
	lLength = ftell( fFile );

	//	Restore the original location within the file.
	fseek( fFile, lCurrentPosition, SEEK_SET);

	//	Return the file's length.
	return lLength;
}

void main(int argc, char **argv)
{
	FILE	*fOriginal, *fNew, *fPatch;
	int		iVerbose = 0;
	char	cByte1, cByte2;
	long	lOriginalLength, lNewLength;
	long	lOffset, lPatchOffset;
	long	lNumberOfDifferences;

	if (argc < 4)
	{
		printf("\nMAKPATCH version 1.0 by Darren Latham\n");
		printf("Use MAKPATCH <filename 1> <filename 2> <patch file> -(q || v)\n");
		printf("-q : Quiet mode, -v : Verbose mode.  Quiet mode is default/\n");
		return;
	}
	else if ( argc == 5 )
	{
		if ( argv[4][1] == 'v' )
			iVerbose = 1;
	}

	fOriginal = fopen( argv[1], "rb" );
	if ( fOriginal == NULL )
	{
		printf("Couldn't open original file <%s>.\n", argv[1] );
		return;
	}

	fNew = fopen( argv[2], "rb+" );
	if ( fNew == NULL )
	{
		printf("Couldn't open changed file <%s>.\n", argv[2] );
		fclose( fOriginal );
		return;
	}

	fPatch = fopen( argv[3], "wb" );
	if ( fPatch == NULL )
	{
		printf("Couldn't open patch file <%s>.\n", argv[3] );
		fclose( fOriginal );
		fclose( fNew );
		return;
	}

	//	Ensure that both of the files are the same length.
	lOriginalLength = filesize( fOriginal );
	lNewLength = filesize( fNew );

	if ( lOriginalLength != lNewLength )
	{
		printf("The files must be same size for a patch to ber produced\n");
		printf("file1 = %lx,  file2 = %lx\n", lOriginalLength, lNewLength );
		fclose( fOriginal );
		fclose( fNew );
		fclose( fPatch );
		return;
	}

	//	Both files are the same length, so, we can produce a patch file.  Write
	//	the length of the files to the patch file.
	fwrite( &lNewLength, sizeof( long ), 1, fPatch );

	//	Move to the beginning of both of the files, and set the current File
	//	Offset counter to 0.
	fseek( fOriginal, 0, SEEK_SET );
	fseek( fNew, 0, SEEK_SET );

	//	While we have not reached the end of the input files
	while ( !feof(fOriginal) )
	{
		//	Read in single bytes from each of the input files, scanning for
		//	differences so we can start a patch record
		cByte1 = (char) fgetc( fOriginal );
		cByte2 = (char) fgetc( fNew );

		if ( cByte1 != cByte2 )
		{
			//	We have found a difference between the two files so record the
			//	current offset in the original/new files within the patch file,
			//	and initialise the number of differences to 1.
			//	We have to subtract 1 from the offset as we have already moved
			//	passed the point where the differences started (the fgetc's)
			lOffset = ftell( fOriginal ) - 1;
			fwrite( &lOffset, sizeof( long ), 1, fPatch );
			lNumberOfDifferences = 0;

			if ( iVerbose ) printf("Start of patch record : %lx\n", lOffset );

			//	Record the offset within the patch file so we can return and
			//	write the actual number of byte differences contained within
			//	this patch record once we know how may there are.  Then, write
			//	a dummy number of differences into the patch file as a place
			//	holder.
			lPatchOffset = ftell( fPatch );
			fwrite( &lNumberOfDifferences, sizeof( long ), 1, fPatch );

			//	For each consecutive different byte pairs found write the
			//	original and new values to the patch file.
			while ( (cByte1 != cByte2) && (!feof(fOriginal)) )
			{
				//	Increae the number of differences in this patch record
				lNumberOfDifferences++;

				fwrite( &cByte1, sizeof( char ), 1, fPatch );
				fwrite( &cByte2, sizeof( char ), 1, fPatch );

				if ( iVerbose ) printf("%2X:%2X  ", cByte1, cByte2);

				//	Get the next pair of values from the patch file.
				cByte1 = (char) fgetc( fOriginal );
				cByte2 = (char) fgetc( fNew );
			}

			//	Now we need to go back to the position in which the number of
			//	differences in this patch record is found and write the true
			//	value to the file.  Then, move back to the end of the patch
			//	file ready to write the next record.
			fseek( fPatch, lPatchOffset, SEEK_SET );
			fwrite( &lNumberOfDifferences, sizeof( long ), 1, fPatch );
			fseek( fPatch, 0, SEEK_END );

			if ( iVerbose ) printf("\nNumber of differences in this record : %ld\n", lNumberOfDifferences );
		}
	}

	fclose( fOriginal );
	fclose( fNew );
	fclose( fPatch );
}
